@@ -1,4 +1,4 @@
-# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2020 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -26,7 +26,8 @@
 # When updating this file, consider if lint.py or warp.py also need
 # updating.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
 
 import renpy.display
 import renpy.test
@@ -89,7 +90,7 @@ class ParameterInfo(object):
             best job it can.
         """
 
-        values = { }
+        values = renpy.python.RevertableDict()
         rv = { }
 
         if args is None:
@@ -107,7 +108,7 @@ class ParameterInfo(object):
 
         extrapos = tuple(args[len(self.positional):])
 
-        for name, value in kwargs.iteritems():
+        for name, value in kwargs.items():
             if name in values:
                 if not ignore_errors:
                     raise Exception("Parameter %s has two values." % name)
@@ -142,7 +143,7 @@ class ParameterInfo(object):
             pass
 
         elif values and (not ignore_errors):
-            raise Exception("Unknown keyword arguments: %s" % ( ", ".join(values.keys())))
+            raise Exception("Unknown keyword arguments: %s" % (", ".join(list(values.keys()))))
 
         return rv
 
@@ -224,7 +225,7 @@ def __newobj__(cls, *args):
 pyexpr_list = [ ]
 
 
-class PyExpr(unicode):
+class PyExpr(str):
     """
     Represents a string containing python code.
     """
@@ -235,7 +236,7 @@ class PyExpr(unicode):
         ]
 
     def __new__(cls, s, filename, linenumber):
-        self = unicode.__new__(cls, s)
+        self = str.__new__(cls, s)
         self.filename = filename
         self.linenumber = linenumber
 
@@ -246,7 +247,7 @@ class PyExpr(unicode):
         return self
 
     def __getnewargs__(self):
-        return (unicode(self), self.filename, self.linenumber)  # E1101
+        return (str(self), self.filename, self.linenumber) # E1101
 
 
 def probably_side_effect_free(expr):
@@ -289,7 +290,7 @@ class PyCode(object):
 
         # The time is necessary so we can disambiguate between Python
         # blocks on the same line in different script versions.
-        self.location = loc + ( int(time.time()), )
+        self.location = loc + (int(time.time()),)
         self.mode = mode
 
         # This will be initialized later on, after we are serialized.
@@ -308,14 +309,14 @@ class PyCode(object):
             pass
 
         code = self.source
-        if isinstance(code, renpy.python.ast.AST):  # @UndefinedVariable
-            code = renpy.python.ast.dump(code)  # @UndefinedVariable
+        if isinstance(code, renpy.python.ast.AST): # @UndefinedVariable
+            code = renpy.python.ast.dump(code) # @UndefinedVariable
 
-        self.hash = chr(renpy.bytecode_version) + hashlib.md5(repr(self.location) + code.encode("utf-8")).digest()
+        self.hash = bchr(renpy.bytecode_version) + hashlib.md5((repr(self.location) + code).encode("utf-8")).digest()
         return self.hash
 
 
-def chain_block(block, next):  # @ReservedAssignment
+def chain_block(block, next): # @ReservedAssignment
     """
     This is called to chain together all of the nodes in a block. Node
     n is chained with node n+1, while the last node is chained with
@@ -341,7 +342,7 @@ class Scry(object):
     def __getattr__(self, name):
         return None
 
-    def next(self):  # @ReservedAssignment
+    def next(self): # @ReservedAssignment
         if self._next is None:
             return None
         else:
@@ -356,9 +357,10 @@ class Node(object):
     A node in the abstract syntax tree of the program.
 
     @ivar name: The name of this node.
-
     @ivar filename: The filename where this node comes from.
     @ivar linenumber: The line number of the line on which this node is defined.
+    @ivar next: The statement that will execute after this one.
+    @ivar statement_start: If present, the first node that makes up the statement that includes this node.
     """
 
     __slots__ = [
@@ -366,6 +368,7 @@ class Node(object):
         'filename',
         'linenumber',
         'next',
+        'statement_start',
         ]
 
     # True if this node is translatable, false otherwise. (This can be set on
@@ -402,7 +405,7 @@ class Node(object):
         tuple returned must be hashable.
         """
 
-        return ( id(self), )
+        return (id(self),)
 
     def get_children(self, f):
         """
@@ -425,7 +428,7 @@ class Node(object):
     # get_init is only present on statements that define it.
     get_init = None
 
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         """
         This is called with the Node node that should be followed after
         executing this node, and all nodes that this node
@@ -464,8 +467,8 @@ class Node(object):
 
     def predict(self):
         """
-        This is called to predictively load images from this node.  It
-        should cause renpy.display.predict.image and
+        This is called to predictively load images from this node. It
+        should cause renpy.display.predict.displayable and
         renpy.display.predict.screen to be called as necessary.
         """
 
@@ -481,7 +484,7 @@ class Node(object):
         """
 
         rv = Scry()
-        rv._next = self.next  # W0201
+        rv._next = self.next # W0201
         return rv
 
     def restructure(self, callback):
@@ -533,7 +536,7 @@ def say_menu_with(expression, callback):
 
     if expression is not None:
         what = renpy.python.py_eval(expression)
-    elif renpy.store.default_transition and renpy.game.preferences.transitions == 2:
+    elif renpy.store.default_transition and renpy.game.preferences.transitions == 2: # @UndefinedVariable
         what = renpy.store.default_transition
     else:
         return
@@ -541,14 +544,11 @@ def say_menu_with(expression, callback):
     if not what:
         return
 
-    if renpy.game.preferences.transitions:
+    if renpy.game.preferences.transitions: # @UndefinedVariable
         # renpy.game.interface.set_transition(what)
         callback(what)
 
 
-fast_who_pattern = re.compile(r'[a-zA-Z_][a-zA-Z_0-9]*$')
-
-
 def eval_who(who, fast=None):
     """
     Evaluates the `who` parameter to a say statement.
@@ -557,25 +557,21 @@ def eval_who(who, fast=None):
     if who is None:
         return None
 
-    if fast is None:
-        fast = bool(fast_who_pattern.match(who))
+    if 'store.character' in renpy.python.store_dicts:
+        rv = renpy.python.store_dicts['store.character'].get(who, None)
+    else:
+        rv = None
 
-    if fast:
-
-        if 'store.character' in renpy.python.store_dicts:
-            rv = renpy.python.store_dicts['store.character'].get(who, None)
-        else:
-            rv = None
-
-        if rv is None:
-            rv = renpy.python.store_dicts['store'].get(who, None)
-
-        if rv is None:
-            raise Exception("Sayer '%s' is not defined." % who.encode("utf-8"))
+    if rv is None:
+        rv = renpy.python.store_dicts['store'].get(who, None)
 
+    if rv is not None:
         return rv
 
-    return renpy.python.py_eval(who)
+    try:
+        return renpy.python.py_eval(who)
+    except:
+        raise Exception("Sayer '%s' is not defined." % who)
 
 
 class Say(Node):
@@ -649,7 +645,7 @@ class Say(Node):
         if dialogue_filter is not None:
             what = dialogue_filter(what)
 
-        rv.append(renpy.translation.encode_say_string(what))
+        rv.append(renpy.translation.encode_say_string(what)) # @UndefinedVariable
 
         if not self.interact:
             rv.append("nointeract")
@@ -666,7 +662,6 @@ class Say(Node):
     def execute(self):
 
         next_node(self.next)
-        statement_name("say")
 
         try:
 
@@ -675,16 +670,21 @@ class Say(Node):
 
             who = eval_who(self.who, self.who_fast)
 
+            if who is not None:
+                statement_name(getattr(who, "statement_name", "say"))
+            else:
+                statement_name("say")
+
             if not (
                     (who is None) or
                     callable(who) or
-                    isinstance(who, basestring) ):
+                    isinstance(who, basestring)):
 
                 raise Exception("Sayer %s is not a function or string." % self.who.encode("utf-8"))
 
             what = self.what
             if renpy.config.say_menu_text_filter:
-                what = renpy.config.say_menu_text_filter(what)  # E1102
+                what = renpy.config.say_menu_text_filter(what) # E1102
 
             renpy.store._last_raw_what = what
 
@@ -717,6 +717,7 @@ class Say(Node):
         try:
 
             renpy.game.context().say_attributes = self.attributes
+            renpy.game.context().temporary_attributes = self.temporary_attributes
 
             who = eval_who(self.who, self.who_fast)
 
@@ -751,7 +752,7 @@ class Say(Node):
 
 
 # Copy the descriptor.
-setattr(Say, "with", Say.with_)  # E1101
+setattr(Say, "with", Say.with_) # E1101
 
 
 class Init(Node):
@@ -779,7 +780,7 @@ class Init(Node):
     # We handle chaining specially. We want to chain together the nodes in
     # the block, but we want that chain to end in None, and we also want
     # this node to just continue on to the next node in normal execution.
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         self.next = next
 
         chain_block(self.block, None)
@@ -836,7 +837,7 @@ class Label(Node):
         for i in self.block:
             i.get_children(f)
 
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
 
         if self.block:
             self.next = self.block[0]
@@ -852,7 +853,7 @@ class Label(Node):
 
         values = apply_arguments(self.parameters, renpy.store._args, renpy.store._kwargs)
 
-        for k, v in values.iteritems():
+        for k, v in values.items():
             renpy.exports.dynamic(k)
             setattr(renpy.store, k, v)
 
@@ -1324,7 +1325,7 @@ class Hide(Node):
 
         layer = renpy.exports.default_layer(layer, tag)
 
-        renpy.game.context().images.predict_hide(tag, layer)
+        renpy.game.context().images.predict_hide(layer, (tag,))
 
         return [ self.next ]
 
@@ -1452,7 +1453,13 @@ class Call(Node):
             if not probably_side_effect_free(label):
                 return [ ]
 
-            label = renpy.python.py_eval(label)
+            try:
+                label = renpy.python.py_eval(label)
+            except:
+                return [ ]
+
+            if not renpy.game.script.has_label(label):
+                return [ ]
 
         return [ renpy.game.context().predict_call(label, self.next.name) ]
 
@@ -1476,10 +1483,10 @@ class Return(Node):
         self.expression = expression
 
     def diff_info(self):
-        return (Return, )
+        return (Return,)
 
     # We don't care what the next node is.
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         self.next = None
         return
 
@@ -1537,7 +1544,7 @@ class Menu(Node):
         self.rollback = "force"
         return self
 
-    def __init__(self, loc, items, set, with_, has_caption, arguments, item_arguments):  # @ReservedAssignment
+    def __init__(self, loc, items, set, with_, has_caption, arguments, item_arguments): # @ReservedAssignment
         super(Menu, self).__init__(loc)
 
         self.items = items
@@ -1559,7 +1566,7 @@ class Menu(Node):
                     i.get_children(f)
 
     # Blocks of statements in a choice continue after the menu.
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
 
         self.next = next
 
@@ -1658,7 +1665,7 @@ class Menu(Node):
                 callback(block)
 
 
-setattr(Menu, "with", Menu.with_)  # E1101
+setattr(Menu, "with", Menu.with_) # E1101
 
 
 # Goto is considered harmful. So we decided to name it "jump"
@@ -1680,7 +1687,7 @@ class Jump(Node):
         return (Jump, self.target, self.expression)
 
     # We don't care what our next node is.
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         self.next = None
         return
 
@@ -1706,7 +1713,13 @@ class Jump(Node):
             if not probably_side_effect_free(label):
                 return [ ]
 
-            label = renpy.python.py_eval(label)
+            try:
+                label = renpy.python.py_eval(label)
+            except:
+                return [ ]
+
+            if not renpy.game.script.has_label(label):
+                return [ ]
 
         return [ renpy.game.script.lookup(label) ]
 
@@ -1755,7 +1768,7 @@ class While(Node):
         for i in self.block:
             i.get_children(f)
 
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         self.next = next
         chain_block(self.block, self)
 
@@ -1808,7 +1821,7 @@ class If(Node):
             for i in block:
                 i.get_children(f)
 
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         self.next = next
 
         for _condition, block in self.entries:
@@ -1895,7 +1908,7 @@ class UserStatement(Node):
             for j in i.block:
                 j.get_children(f)
 
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         self.next = next
 
         if self.code_block is not None:
@@ -2046,6 +2059,9 @@ class StoreNamespace(object):
     def set(self, name, value):
         renpy.python.store_dicts[self.store][name] = value
 
+    def get(self, name):
+        return renpy.python.store_dicts[self.store][name]
+
 
 def get_namespace(store):
     """
@@ -2061,7 +2077,7 @@ def get_namespace(store):
 
 # Config variables that are set twice - once when the rpy is first loaded,
 # and then again at init time.
-EARLY_CONFIG = { "save_directory", "allow_duplicate_labels", "keyword_after_python" }
+EARLY_CONFIG = { "save_directory", "allow_duplicate_labels", "keyword_after_python", "steam_appid" }
 
 define_statements = [ ]
 
@@ -2072,18 +2088,29 @@ class Define(Node):
         'varname',
         'code',
         'store',
+        'operator',
+        'index',
         ]
 
     def __new__(cls, *args, **kwargs):
         self = Node.__new__(cls)
         self.store = 'store'
+        self.operator = '='
+        self.index = None
         return self
 
-    def __init__(self, loc, store, name, expr):
+    def __init__(self, loc, store, name, index, operator, expr):
         super(Define, self).__init__(loc)
 
         self.store = store
         self.varname = name
+
+        if index is not None:
+            self.index = PyCode(index, loc=loc, mode='eval')
+        else:
+            self.index = None
+
+        self.operator = operator
         self.code = PyCode(expr, loc=loc, mode='eval')
 
     def diff_info(self):
@@ -2092,7 +2119,14 @@ class Define(Node):
     def early_execute(self):
         create_store(self.store)
 
+        if self.operator != "=":
+            return
+
+        if self.index is not None:
+            return
+
         if self.store == "store.config" and self.varname in EARLY_CONFIG:
+
             value = renpy.python.py_eval_bytecode(self.code.bytecode)
             setattr(renpy.config, self.varname, value)
 
@@ -2103,25 +2137,52 @@ class Define(Node):
 
         define_statements.append(self)
 
-        value = renpy.python.py_eval_bytecode(self.code.bytecode)
-
         if self.store == 'store':
-            renpy.exports.pure(self.varname)
             renpy.dump.definitions.append((self.varname, self.filename, self.linenumber))
         else:
             renpy.dump.definitions.append((self.store[6:] + "." + self.varname, self.filename, self.linenumber))
 
-        ns, _special = get_namespace(self.store)
-        ns.set(self.varname, value)
+        if self.operator == "=" and self.index is None:
+            renpy.exports.pure(self.store + "." + self.varname)
+
+        self.set()
 
     def redefine(self, stores):
 
         if self.store not in stores:
             return
 
+        self.set()
+
+    def set(self):
+
         value = renpy.python.py_eval_bytecode(self.code.bytecode)
         ns, _special = get_namespace(self.store)
-        ns.set(self.varname, value)
+
+        if (self.index is None) and (self.operator == "="):
+            ns.set(self.varname, value)
+            return
+
+        base = ns.get(self.varname)
+        old = base
+
+        if self.index:
+            key = renpy.python.py_eval_bytecode(self.index.bytecode)
+
+            if self.operator != "=":
+                old = base[key]
+
+        if self.operator == "=":
+            new = value
+        elif self.operator == "+=":
+            new = old + value
+        elif self.operator == "|=":
+            new = old | value
+
+        if self.index:
+            base[key] = new
+        else:
+            ns.set(self.varname, new)
 
 
 def redefine(stores):
@@ -2133,7 +2194,6 @@ def redefine(stores):
         i.redefine(stores)
 
 
-
 # All the default statements, in the order they were registered.
 default_statements = [ ]
 
@@ -2220,10 +2280,9 @@ class Screen(Node):
 
     def __init__(self, loc, screen):
         """
-        @param name: The name of the image being defined.
-
-        @param expr: An expression yielding a Displayable that is
-        assigned to the image.
+        @param screen: The screen object being defined.
+        In SL1, an instance of screenlang.ScreenLangScreen.
+        In SL2, an instance of sl2.slast.SLScreen.
         """
 
         super(Screen, self).__init__(loc)
@@ -2240,11 +2299,11 @@ class Screen(Node):
         self.screen.define((self.filename, self.linenumber))
         renpy.dump.screens.append((self.screen.name, self.filename, self.linenumber))
 
-
 ################################################################################
 # Translations
 ################################################################################
 
+
 class Translate(Node):
     """
     A translation block, produced either by explicit translation statements
@@ -2282,7 +2341,7 @@ class Translate(Node):
     def diff_info(self):
         return (Translate, self.identifier, self.language)
 
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         if self.block:
             self.next = self.block[0]
             chain_block(self.block, next)
@@ -2311,8 +2370,8 @@ class Translate(Node):
             next_node(self.next)
             raise Exception("Translation nodes cannot be run directly.")
 
-        if self.identifier not in renpy.game.persistent._seen_translates:  # @UndefinedVariable
-            renpy.game.persistent._seen_translates.add(self.identifier)  # @UndefinedVariable
+        if self.identifier not in renpy.game.persistent._seen_translates: # @UndefinedVariable
+            renpy.game.persistent._seen_translates.add(self.identifier) # @UndefinedVariable
             renpy.game.seen_translates_count += 1
             renpy.game.new_translates_count += 1
 
@@ -2461,7 +2520,7 @@ class TranslateBlock(Node):
     # We handle chaining specially. We want to chain together the nodes in
     # the block, but we want that chain to end in None, and we also want
     # this node to just continue on to the next node in normal execution.
-    def chain(self, next):  # @ReservedAssignment
+    def chain(self, next): # @ReservedAssignment
         self.next = next
         chain_block(self.block, None)
 
@@ -2529,7 +2588,7 @@ class Style(Node):
             if not renpy.exports.variant(variant):
                 return
 
-        s = renpy.style.get_or_create_style(self.style_name)  # @UndefinedVariable
+        s = renpy.style.get_or_create_style(self.style_name) # @UndefinedVariable
 
         if self.clear:
             s.clear()
@@ -2589,3 +2648,26 @@ class Testcase(Node):
         statement_name("testcase")
 
         renpy.test.testexecution.testcases[self.label] = self.test
+
+
+class RPY(Node):
+    __slots__ = [
+        "rest"
+        ]
+
+    def __init__(self, loc, rest):
+        super(RPY, self).__init__(loc)
+
+        self.rest = rest
+
+    def diff_info(self):
+        return (RPY, self.rest)
+
+    def execute(self):
+        next_node(self.next)
+        statement_name("rpy")
+
+        # rpy python 3 is run in Script.finish_load.
+
+    def get_code(self):
+        return "rpy " + " ".join(self.rest)
