@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2020 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,6 +19,9 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+
 import renpy.display
 import renpy.text
 import codecs
@@ -27,10 +30,9 @@ import re
 import sys
 import collections
 import textwrap
+import builtins
 
-import __builtin__
-
-python_builtins = set(dir(__builtin__))
+python_builtins = set(dir(builtins))
 renpy_builtins = set()
 
 image_prefixes = None
@@ -50,7 +52,13 @@ image_prefixes = None
 # The node the report will be about:
 report_node = None
 
+# Collect define/default statements to check for duplication
+all_define_statments = {}
+all_default_statements = {}
+
 # Reports a message to the user.
+
+
 def report(msg, *args):
     if report_node:
         out = u"%s:%d " % (renpy.parser.unicode_filename(report_node.filename), report_node.linenumber)
@@ -58,21 +66,43 @@ def report(msg, *args):
         out = ""
 
     out += msg % args
-    print
-    print out.encode('utf-8')
+    print()
+    print(out.encode('utf-8'))
+
 
 added = { }
 
 # Reports additional information about a message, the first time it
 # occurs.
-def add(msg):
+
+
+def add(msg, *args):
     if not msg in added:
         added[msg] = True
-        print unicode(msg).encode('utf-8')
+        msg = str(msg) % args
+        print(msg)
 
 
-# Trys to evaluate an expression, announcing an error if it fails.
+# Tries to evaluate an expression, announcing an error if it fails.
 def try_eval(where, expr, additional=None):
+    """
+    :doc: lint
+
+    Tries to evaluate an expression, and writes an error to lint.txt if
+    it fails.
+
+    `where`
+        A string giving the location the expression is found. Used to
+        generate an error message of the form "Could not evaluate `expr`
+        in `where`."
+
+    `expr`
+        The expression to try evaluating.
+
+    `additional`
+        If given, an additional line of information that is addded to the
+        error message.
+    """
 
     # Make sure the expression compiles.
     try_compile(where, expr)
@@ -90,13 +120,33 @@ def try_eval(where, expr, additional=None):
     if m.group(1) in __builtins__:
         return
 
-    report( "Could not evaluate '%s', in %s.", expr, where)
+    report("Could not evaluate '%s', in %s.", expr, where)
     if additional:
         add(additional)
 
 # Returns True of the expression can be compiled as python, False
 # otherwise.
+
+
 def try_compile(where, expr, additional=None):
+    """
+    :doc: lint
+
+    Tries to compile an expression, and writes an error to lint.txt if
+    it fails.
+
+    `where`
+        A string giving the location the expression is found. Used to
+        generate an error message of the form "Could not evaluate `expr`
+        in `where`."
+
+    `expr`
+        The expression to try compiling.
+
+    `additional`
+        If given, an additional line of information that is addded to the
+        error message.
+    """
 
     try:
         renpy.python.py_compile_eval_bytecode(expr)
@@ -109,6 +159,7 @@ def try_compile(where, expr, additional=None):
 # The sets of names + attributes that we know are valid.
 imprecise_cache = set()
 
+
 def image_exists_imprecise(name):
     """
     Returns true if the image is a plausible image that can be used in a show
@@ -116,7 +167,6 @@ def image_exists_imprecise(name):
     tag and containing all of the attributes (and none of the removed attributes).
     """
 
-
     if name in imprecise_cache:
         return True
 
@@ -131,17 +181,22 @@ def image_exists_imprecise(name):
         else:
             required.add(i)
 
-    for im in renpy.display.image.images:
+    for im, d in renpy.display.image.images.items():
 
         if im[0] != nametag:
             continue
 
         attrs = set(im[1:])
 
-        if [ i for i in required if i not in attrs ]:
+        if [ i for i in banned if i in attrs ]:
             continue
 
-        if [ i for i in banned if i in attrs ]:
+        li = getattr(d, "_list_attributes", None)
+
+        if li is not None:
+            attrs = attrs | set(li(im[0], required))
+
+        if [ i for i in required if i not in attrs ]:
             continue
 
         imprecise_cache.add(name)
@@ -152,6 +207,7 @@ def image_exists_imprecise(name):
 
 precise_cache = set()
 
+
 def image_exists_precise(name):
     """
     Returns true if an image exists with the same tag and attributes as
@@ -163,18 +219,41 @@ def image_exists_precise(name):
 
     nametag = name[0]
 
-    required = set(name[1:])
+    required = set()
+    banned = set()
 
-    for im in renpy.display.image.images:
+    for i in name[1:]:
+        if i[0] == "-":
+            banned.add(i[1:])
+        else:
+            required.add(i)
+
+    for im, d in renpy.display.image.images.items():
 
         if im[0] != nametag:
             continue
 
         attrs = set(im[1:])
 
-        if attrs == required:
-            precise_cache.add(name)
-            return True
+        if attrs - required:
+            continue
+
+        rest = required - attrs
+
+        if rest:
+
+            try:
+                da = renpy.display.core.DisplayableArguments()
+                da.name = (im[0],) + tuple(i for i in name[1:] if i in attrs)
+                da.args = tuple(i for i in name[1:] if i in rest)
+                da.lint = True
+                d._duplicate(da)
+            except:
+                continue
+
+        precise_cache.add(name)
+
+        return True
 
     return False
 
@@ -193,30 +272,22 @@ def image_exists(name, expression, tag, precise=True):
     if expression:
         return
 
-    namelist = list(name)
-    names = " ".join(namelist)
-
-    # Look for the precise name.
-    while namelist:
-        if tuple(namelist) in renpy.display.image.images:
-            return
-
-        namelist.pop()
-
-    # If we're not precise, then we have to start looking for images
-    # that we can possibly match.
-    if precise:
-        if image_exists_precise(name):
-            return
-    else:
+    if not precise:
         if image_exists_imprecise(name):
             return
 
-    report("The image named '%s' was not declared.", names)
+    # If we're not precise, then we have to start looking for images
+    # that we can possibly match.
+    if image_exists_precise(name):
+        return
+
+    report("'%s' is not an image.", " ".join(name))
+
 
 # Only check each file once.
 check_file_cache = { }
 
+
 def check_file(what, fn):
 
     present = check_file_cache.get(fn, None)
@@ -245,7 +316,7 @@ def check_displayable(what, d):
 
     try:
         if isinstance(d, renpy.display.core.Displayable):
-            d.visit_all(lambda a : a.predict_one())
+            d.visit_all(lambda a: a.predict_one())
     except:
         pass
 
@@ -260,6 +331,7 @@ def check_image(node):
 
     check_displayable('image %s' % name, renpy.display.image.images[node.imgname])
 
+
 def imspec(t):
     if len(t) == 3:
         return t[0], None, None, t[1], t[2], 0
@@ -297,9 +369,9 @@ def precheck_show(node):
     tag = imspec(node.imspec)[2]
     image_prefixes[tag] = True
 
-
 # Lints ast.Hide.
 
+
 def check_hide(node):
 
     name, _expression, tag, _at_list, layer, _zorder, _behind = imspec(node.imspec)
@@ -318,6 +390,7 @@ def check_hide(node):
 def check_with(node):
     try_eval("a with statement or clause", node.expr, "Perhaps you forgot to declare, or misspelled, a transition?")
 
+
 def check_user(node):
 
     def error(msg):
@@ -334,6 +407,7 @@ def check_user(node):
     except:
         report("Didn't properly report what the next statement should be.")
 
+
 def text_checks(s):
     msg = renpy.text.extras.check_text_tags(s)
     if msg:
@@ -376,6 +450,7 @@ def text_checks(s):
         if state != 0:
             report("Unterminated string format code '%s' (in %s)", fmt, repr(s)[1:])
 
+
 def check_say(node):
 
     if node.who:
@@ -412,7 +487,7 @@ def check_say(node):
     if image_exists_imprecise(name):
         return
 
-    if image_exists_imprecise(('side', ) + name):
+    if image_exists_imprecise(('side',) + name):
         return
 
     report("Could not find image (%s) corresponding to attributes on say statement.", " ".join(name))
@@ -432,6 +507,7 @@ def check_menu(node):
 
         text_checks(l)
 
+
 def check_jump(node):
 
     if node.expression:
@@ -440,6 +516,7 @@ def check_jump(node):
     if not renpy.game.script.has_label(node.target):
         report("The jump is to nonexistent label '%s'.", node.target)
 
+
 def check_call(node):
 
     if node.expression:
@@ -448,14 +525,16 @@ def check_call(node):
     if not renpy.game.script.has_label(node.label):
         report("The call is to nonexistent label '%s'.", node.label)
 
+
 def check_while(node):
     try_compile("in the condition of the while statement", node.condition)
 
-def check_if(node):
 
+def check_if(node):
     for condition, _block in node.entries:
         try_compile("in a condition of the if statement", condition)
 
+
 def check_define(node, kind):
     if node.store != 'store':
         return
@@ -470,6 +549,45 @@ def check_define(node, kind):
         report("'%s %s' replaces a Ren'Py built-in name, which may cause problems.", kind, node.varname)
 
 
+def check_redefined(node, kind):
+    """
+    Check if a define or default statement has already been created.
+    """
+    if kind == 'default':
+        scanned = all_default_statements
+    elif kind == 'define':
+        scanned = all_define_statments
+
+        if not (node.operator == "=" and node.index is None):
+            return
+
+    # Combine store name and varname
+
+    store_name = node.store
+    if store_name.startswith("store."):
+        store_name = store_name[6:]
+
+    if store_name:
+        full_name = "{}.{}".format(store_name, node.varname)
+    else:
+        full_name = node.varname
+
+    if full_name in renpy.config.lint_ignore_redefine:
+        return
+
+    original_node = scanned.get(full_name)
+    if original_node:
+        report(
+            "{} {} already defined at {}:{}".format(
+                kind,
+                full_name,
+                original_node.filename,
+                original_node.linenumber,
+            )
+        )
+    scanned[full_name] = node
+
+
 def check_style_property_displayable(name, property, d):
 
     if not d._duplicatable:
@@ -482,7 +600,7 @@ def check_style_property_displayable(name, property, d):
 
     def sort_short(l):
         l = list(l)
-        l.sort(key=lambda a : len(a))
+        l.sort(key=lambda a: len(a))
         return l
 
     alts = sort_short(renpy.style.prefix_alts)
@@ -493,6 +611,7 @@ def check_style_property_displayable(name, property, d):
             if rest in renpy.style.all_properties:
                 args = d._args.copy(prefix=prefix)
                 dd = d._duplicate(args)
+                dd._unique()
 
                 check_displayable(
                     "{}, property {}".format(name, prefix + property),
@@ -506,15 +625,19 @@ def check_style_property_displayable(name, property, d):
 def check_style(name, s):
 
     for p in s.properties:
-        for k, v in p.iteritems():
+        for k, v in p.items():
 
             # Treat font specially.
             if k.endswith("font"):
-                check_file(name, v)
+                if isinstance(v, renpy.text.font.FontGroup):
+                    for f in set(v.map.values()):
+                        check_file(name, f)
+                else:
+                    check_file(name, v)
 
             if isinstance(v, renpy.display.core.Displayable):
                 check_style_property_displayable(name, k, v)
-#                check_displayable(kname, v)
+
 
 def check_label(node):
 
@@ -535,14 +658,22 @@ def check_label(node):
         add_arg(pi.extrakw)
 
 
+def check_screen(node):
+
+    if (node.screen.parameters is None) and renpy.config.lint_screens_without_parameters:
+        report("The screen %s has not been given a parameter list.", node.screen.name)
+        add("This can be fixed by writing 'screen %s():' instead.", node.screen.name)
+
+
 def check_styles():
-    for full_name, s in renpy.style.styles.iteritems(): # @UndefinedVariable
+    for full_name, s in renpy.style.styles.items(): # @UndefinedVariable
         name = "style." + full_name[0]
         for i in full_name[1:]:
             name += "[{!r}]".format(i)
 
         check_style("Style " + name, s)
 
+
 def humanize(n):
     s = str(n)
 
@@ -556,6 +687,7 @@ def humanize(n):
 
     return ''.join(rv)
 
+
 def check_filename_encodings():
     """
     Checks files to ensure that they are displayable in unicode.
@@ -571,6 +703,7 @@ def check_filename_encodings():
         report("%s contains non-ASCII characters in its filename.", filename)
         add("(ZIP file distributions can only reliably include ASCII filenames.)")
 
+
 class Count(object):
     """
     Stores information about the word count.
@@ -591,6 +724,7 @@ class Count(object):
         self.words += len(s.split())
         self.characters += len(s)
 
+
 def common(n):
     """
     Returns true if the node is in the common directory.
@@ -603,6 +737,7 @@ def common(n):
     else:
         return False
 
+
 def lint():
     """
     The master lint function, that's responsible for staging all of the
@@ -615,13 +750,12 @@ def lint():
     args = ap.parse_args()
 
     if args.filename:
-        f = open(args.filename, "w")
+        f = codecs.open(args.filename, "w", encoding="utf-8")
         sys.stdout = f
 
     renpy.game.lint = True
 
-    print codecs.BOM_UTF8
-    print unicode(renpy.version + " lint report, generated at: " + time.ctime()).encode("utf-8")
+    print("\ufeff" + renpy.version + " lint report, generated at: " + time.ctime())
 
     # This supports check_hide.
     global image_prefixes
@@ -633,8 +767,8 @@ def lint():
     # Iterate through every statement in the program, processing
     # them. We sort them in filename, linenumber order.
 
-    all_stmts = [ (i.filename, i.linenumber, i) for i in renpy.game.script.all_stmts ]
-    all_stmts.sort()
+    all_stmts = list(renpy.game.script.all_stmts)
+    all_stmts.sort(key=lambda n : (n.filename, n.linenumber))
 
     # The current count.
     counts = collections.defaultdict(Count)
@@ -648,11 +782,11 @@ def lint():
 
     global report_node
 
-    for _fn, _ln, node in all_stmts:
+    for node in all_stmts:
         if isinstance(node, (renpy.ast.Show, renpy.ast.Scene)):
             precheck_show(node)
 
-    for _fn, _ln, node in all_stmts:
+    for node in all_stmts:
 
         if common(node):
             continue
@@ -680,7 +814,6 @@ def lint():
 
             counts[language].add(node.what)
 
-
         elif isinstance(node, renpy.ast.Menu):
             check_menu(node)
             menu_count += 1
@@ -711,14 +844,15 @@ def lint():
 
         elif isinstance(node, renpy.ast.Screen):
             screen_count += 1
+            check_screen(node)
 
         elif isinstance(node, renpy.ast.Define):
             check_define(node, "define")
+            check_redefined(node, "define")
 
         elif isinstance(node, renpy.ast.Default):
             check_define(node, "default")
-
-
+            check_redefined(node, "default")
 
     report_node = None
 
@@ -728,7 +862,6 @@ def lint():
     for f in renpy.config.lint_hooks:
         f()
 
-
     lines = [ ]
 
     def report_language(language):
@@ -754,11 +887,10 @@ characters per block. """.format(
 
         lines.append(s)
 
-
-    print
-    print
-    print "Statistics:"
-    print
+    print()
+    print()
+    print("Statistics:")
+    print()
 
     languages = list(counts)
     languages.sort()
@@ -770,17 +902,19 @@ characters per block. """.format(
 
     for l in lines:
         for ll in textwrap.wrap(l, 78):
-            print ll.encode("utf-8")
+            print(ll.encode("utf-8"))
 
-        print
+        print()
 
-    print
-    if renpy.config.developer and (renpy.config.developer != "auto"):
-        print "Remember to set config.developer to False before releasing."
-        print
+    for i in renpy.config.lint_stats_callbacks:
+        i()
 
-    print "Lint is not a substitute for thorough testing. Remember to update Ren'Py"
-    print "before releasing. New releases fix bugs and improve compatibility."
+    print()
+    if renpy.config.developer and (renpy.config.original_developer != "auto"):
+        print("Remember to set config.developer to False before releasing.")
+        print()
+
+    print("Lint is not a substitute for thorough testing. Remember to update Ren'Py")
+    print("before releasing. New releases fix bugs and improve compatibility.")
 
     return False
-
