@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2020 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,6 +19,9 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+
 import renpy.display
 import renpy.style
 import renpy.sl2
@@ -30,10 +33,14 @@ import os
 import sys
 import time
 import zipfile
+import gc
+import linecache
+
 import __main__
 
 last_clock = time.time()
 
+
 def log_clock(s):
     global last_clock
     now = time.time()
@@ -41,10 +48,15 @@ def log_clock(s):
 
     renpy.display.log.write(s)
     if renpy.android and not renpy.config.log_to_stdout:
-        print s
+        print(s)
+
+    # Pump the presplash window to prevent marking
+    # our process as unresponsive by OS
+    renpy.display.presplash.pump_window()
 
     last_clock = now
 
+
 def reset_clock():
     global last_clock
     last_clock = time.time()
@@ -95,7 +107,9 @@ def run(restart):
 
     # Handle arguments and commands.
     if not renpy.arguments.post_init():
-        renpy.exports.quit()
+        # We use 'exception' instead of exports.quit
+        # to not call quit label since it's not nessesary.
+        raise renpy.game.QuitException()
 
     if renpy.config.clear_lines:
         renpy.scriptedit.lines.clear()
@@ -117,16 +131,6 @@ def run(restart):
 
     game.context().goto_label(start_label)
 
-    # Perhaps warp.
-    warp_label = renpy.warp.warp()
-
-    if warp_label is not None:
-
-        game.context().goto_label(warp_label)
-        game.context().call('_after_warp')
-
-        renpy.config.skipping = None
-
     try:
         renpy.exports.log("--- " + time.ctime())
         renpy.exports.log("")
@@ -146,12 +150,12 @@ def run(restart):
 
 def load_rpe(fn):
 
-    zfn = zipfile.ZipFile(fn)
-    autorun = zfn.read("autorun.py")
-    zfn.close()
+    with zipfile.ZipFile(fn) as zfn:
+        autorun = zfn.read("autorun.py")
 
     sys.path.insert(0, fn)
-    exec autorun in dict()
+    exec(autorun, dict())
+
 
 def choose_variants():
 
@@ -161,17 +165,16 @@ def choose_variants():
 
     renpy.config.variants = [ None ]
 
-    if renpy.android: #@UndefinedVariable
+    if renpy.android: # @UndefinedVariable
 
         renpy.config.variants.insert(0, 'mobile')
         renpy.config.variants.insert(0, 'android')
 
-        import android #@UnresolvedImport
+        import android # @UnresolvedImport
         import math
         import pygame_sdl2 as pygame
 
-        from jnius import autoclass  # @UnresolvedImport
-
+        from jnius import autoclass # @UnresolvedImport
 
         # Manufacturer/Model-specific variants.
         try:
@@ -180,10 +183,10 @@ def choose_variants():
             manufacturer = Build.MANUFACTURER
             model = Build.MODEL
 
-            print "Manufacturer", manufacturer, "model", model
+            print("Manufacturer", manufacturer, "model", model)
 
             if manufacturer == "Amazon" and model.startswith("AFT"):
-                print "Running on a Fire TV."
+                print("Running on a Fire TV.")
                 renpy.config.variants.insert(0, "firetv")
         except:
             pass
@@ -192,7 +195,7 @@ def choose_variants():
         package_manager = android.activity.getPackageManager()
 
         if package_manager.hasSystemFeature("android.hardware.type.television"):
-            print "Running on a television."
+            print("Running on a television.")
             renpy.config.variants.insert(0, "tv")
             renpy.config.variants.insert(0, "small")
             return
@@ -204,7 +207,7 @@ def choose_variants():
 
         info = renpy.display.get_info()
         diag = math.hypot(info.current_w, info.current_h) / android.get_dpi()
-        print "Screen diagonal is", diag, "inches."
+        print("Screen diagonal is", diag, "inches.")
 
         if diag >= 6:
             renpy.config.variants.insert(0, 'tablet')
@@ -214,6 +217,7 @@ def choose_variants():
             renpy.config.variants.insert(0, 'small')
 
     elif renpy.ios:
+        renpy.config.variants.insert(0, 'mobile')
         renpy.config.variants.insert(0, 'ios')
         renpy.config.variants.insert(0, 'touch')
 
@@ -222,7 +226,7 @@ def choose_variants():
 
         idiom = UIDevice.currentDevice().userInterfaceIdiom
 
-        print "iOS device idiom", idiom
+        print("iOS device idiom", idiom)
 
         # idiom 0 is iPhone, 1 is iPad. We assume any bigger idiom will
         # be tablet-like.
@@ -233,18 +237,72 @@ def choose_variants():
             renpy.config.variants.insert(0, 'phone')
             renpy.config.variants.insert(0, 'small')
 
+    elif renpy.emscripten:
+        import emscripten
+        import re
+
+        # web
+        renpy.config.variants.insert(0, 'web')
+
+        # mobile
+        userAgent = emscripten.run_script_string(r'''navigator.userAgent''')
+        mobile = re.search('Mobile|Android|iPad|iPhone', userAgent)
+        if mobile:
+            renpy.config.variants.insert(0, 'mobile')
+        # Reserve android/ios for when the OS API is exposed
+        # if re.search('Android', userAgent):
+        #    renpy.config.variants.insert(0, 'android')
+        # if re.search('iPad|iPhone', userAgent):
+        #    renpy.config.variants.insert(0, 'ios')
+
+        # touch
+        touch = emscripten.run_script_int(r'''
+          ('ontouchstart' in window) ||
+            (navigator.maxTouchPoints > 0) ||
+            (navigator.msMaxTouchPoints > 0)''')
+        if touch == 1:
+            # mitigate hybrids (e.g. ms surface) by restricting touch to mobile
+            if mobile:
+                renpy.config.variants.insert(0, 'touch')
+
+        # large/medium/small
+        # tablet/phone
+        # screen.width/height is auto-adjusted by browser,
+        # so it can be used as a physical sizereference
+        # (see also window.devicePixelRatio)
+        # e.g. Galaxy S5:
+        # - physical / OpenGL: 1080x1920
+        # - web screen: 360x640 w/ devicePixelRatio=3
+        ref_width = emscripten.run_script_int(r'''screen.width''')
+        ref_height = emscripten.run_script_int(r'''screen.height''')
+        # medium reference point: ipad 1024x768, ipad pro 1336x1024 (browser "pixels")
+        if mobile:
+            if (ref_width < 768 or ref_height < 768):
+                renpy.config.variants.insert(0, 'small')
+                renpy.config.variants.insert(0, 'phone')
+            else:
+                renpy.config.variants.insert(0, 'medium')
+                renpy.config.variants.insert(0, 'tablet')
+        else:
+            renpy.config.variants.insert(0, 'large')
+
     else:
         renpy.config.variants.insert(0, 'pc')
-        renpy.config.variants.insert(0, 'large')
 
+        renpy.config.variants.insert(0, 'large')
 
 
 def main():
 
+    gc.set_threshold(*renpy.config.gc_thresholds)
+
     log_clock("Bootstrap to the start of init.init")
 
     renpy.game.exception_info = 'Before loading the script.'
 
+    # Clear the line cache, since the script may have changed.
+    linecache.clearcache()
+
     # Get ready to accept new arguments.
     renpy.arguments.pre_init()
 
@@ -254,6 +312,12 @@ def main():
     # Init the config after load.
     renpy.config.init()
 
+    # Reset live2d if it exists.
+    try:
+        renpy.gl2.live2d.reset()
+    except:
+        pass
+
     # Set up variants.
     choose_variants()
     renpy.display.touch = "touch" in renpy.config.variants
@@ -273,25 +337,44 @@ def main():
     else:
         renpy.config.commondir = None
 
+    # Add path from env variable, if any
+    if "RENPY_SEARCHPATH" in os.environ:
+        renpy.config.searchpath.extend(os.environ["RENPY_SEARCHPATH"].split("::"))
+
     if renpy.android:
         renpy.config.searchpath = [ ]
         renpy.config.commondir = None
 
+        if "ANDROID_PUBLIC" in os.environ:
+            android_game = os.path.join(os.environ["ANDROID_PUBLIC"], "game")
+
+            print("Android searchpath: ", android_game)
+
+            if os.path.exists(android_game):
+                renpy.config.searchpath.insert(0, android_game)
+
     # Load Ren'Py extensions.
-    for dir in renpy.config.searchpath: #@ReservedAssignment
+    for dir in renpy.config.searchpath: # @ReservedAssignment
         for fn in os.listdir(dir):
             if fn.lower().endswith(".rpe"):
                 load_rpe(dir + "/" + fn)
 
+    # Generate a list of extensions for each archive handler.
+    archive_extensions = [ ]
+    for handler in renpy.loader.archive_handlers:
+        for ext in handler.get_supported_extensions():
+            if not (ext in archive_extensions):
+                archive_extensions.append(ext)
 
     # The basename is the final component of the path to the gamedir.
     for i in sorted(os.listdir(renpy.config.gamedir)):
+        base, ext = os.path.splitext(i)
 
-        if not i.endswith(".rpa"):
+        # Check if the archive does not have any of the extensions in archive_extensions
+        if not (ext in archive_extensions):
             continue
 
-        i = i[:-4]
-        renpy.config.archives.append(i)
+        renpy.config.archives.append(base)
 
     renpy.config.archives.reverse()
 
@@ -328,6 +411,11 @@ def main():
 
     # Set up error handling.
     renpy.exports.load_module("_errorhandling")
+
+    if renpy.exports.loadable("tl/None/common.rpym") or renpy.exports.loadable("tl/None/common.rpymc"):
+        renpy.exports.load_module("tl/None/common")
+
+    renpy.config.init_system_styles()
     renpy.style.build_styles() # @UndefinedVariable
 
     log_clock("Loading error handling")
@@ -337,15 +425,16 @@ def main():
     # labels as in other scripts (usually happens on script rename).
     if (renpy.game.args.command == 'compile') and not (renpy.game.args.keep_orphan_rpyc): # @UndefinedVariable
 
-        for (fn, _dir) in renpy.game.script.script_files:
+        for (fn, dn) in renpy.game.script.script_files:
 
-            if dir is None:
+            if dn is None:
                 continue
 
-            if not os.path.isfile(os.path.join(dir, fn+".rpy")):
+            if not os.path.isfile(os.path.join(dn, fn + ".rpy")):
+
                 try:
-                    name = os.path.join(dir, fn+".rpyc")
-                    os.rename(name, name+".bak")
+                    name = os.path.join(dn, fn + ".rpyc")
+                    os.rename(name, name + ".bak")
                 except OSError:
                     # This perhaps shouldn't happen since either .rpy or .rpyc should exist
                     pass
@@ -366,7 +455,7 @@ def main():
             renpy.game.script = renpy.script.Script()
             renpy.game.script.load_script()
 
-        print time.time() - start
+        print(time.time() - start)
         sys.exit(0)
 
     renpy.game.exception_info = 'After loading the script.'
@@ -375,24 +464,34 @@ def main():
     if renpy.config.savedir is None:
         renpy.config.savedir = __main__.path_to_saves(renpy.config.gamedir) # E1101 @UndefinedVariable
 
-    if renpy.game.args.savedir: #@UndefinedVariable
-        renpy.config.savedir = renpy.game.args.savedir #@UndefinedVariable
+    if renpy.game.args.savedir: # @UndefinedVariable
+        renpy.config.savedir = renpy.game.args.savedir # @UndefinedVariable
 
     # Init preferences.
     game.persistent = renpy.persistent.init()
     game.preferences = game.persistent._preferences
 
+    for i in renpy.game.persistent._seen_translates: # @UndefinedVariable
+        if i in renpy.game.script.translator.default_translates:
+            renpy.game.seen_translates_count += 1
+
     if game.persistent._virtual_size:
         renpy.config.screen_width, renpy.config.screen_height = game.persistent._virtual_size
 
-    # Init save locations.
+    # Init save locations and loadsave.
     renpy.savelocation.init()
 
     # We need to be 100% sure we kill the savelocation thread.
     try:
 
+        # Init save slots.
+        renpy.loadsave.init()
+
+        log_clock("Loading save slot metadata.")
+
         # Load persistent data from all save locations.
         renpy.persistent.update()
+        game.preferences = game.persistent._preferences
         log_clock("Loading persistent")
 
         # Clear the list of seen statements in this game.
@@ -409,22 +508,39 @@ def main():
         renpy.game.exception_info = 'While executing init code:'
 
         for _prio, node in game.script.initcode:
-            game.context().run(node)
+
+            if isinstance(node, renpy.ast.Node):
+                node_start = time.time()
+
+                renpy.game.context().run(node)
+
+                node_duration = time.time() - node_start
+
+                if node_duration > renpy.config.profile_init:
+                    renpy.display.log.write(" - Init at %s:%d took %.5f s.", node.filename, node.linenumber, node_duration)
+
+            else:
+                # An init function.
+                node()
 
         renpy.game.exception_info = 'After initialization, but before game start.'
 
         # Check if we should simulate android.
-        renpy.android = renpy.android or renpy.config.simulate_android #@UndefinedVariable
+        renpy.android = renpy.android or renpy.config.simulate_android # @UndefinedVariable
+
+        # Re-set up the logging.
+        renpy.log.post_init()
 
         # Run the post init code, if any.
         for i in renpy.game.post_init:
             i()
 
-        game.persistent._virtual_size = renpy.config.screen_width, renpy.config.screen_height
+        renpy.game.script.report_duplicate_labels()
 
-        for i in renpy.game.persistent._seen_translates: # @UndefinedVariable
-            if i in renpy.game.script.translator.default_translates:
-                renpy.game.seen_translates_count += 1
+        # Sort the images.
+        renpy.display.image.image_names.sort()
+
+        game.persistent._virtual_size = renpy.config.screen_width, renpy.config.screen_height
 
         log_clock("Running init code")
 
@@ -460,6 +576,29 @@ def main():
         renpy.python.make_clean_stores()
         log_clock("Making clean stores")
 
+        gc.collect(2)
+
+        if gc.garbage:
+            del gc.garbage[:]
+
+        if renpy.config.manage_gc:
+            gc.set_threshold(*renpy.config.gc_thresholds)
+
+            gc_debug = int(os.environ.get("RENPY_GC_DEBUG", 0))
+
+            if renpy.config.gc_print_unreachable:
+                gc_debug |= gc.DEBUG_SAVEALL
+
+            gc.set_debug(gc_debug)
+
+        else:
+            gc.set_threshold(700, 10, 10)
+
+        log_clock("Initial gc.")
+
+        # Start debugging file opens.
+        renpy.debug.init_main_thread_open()
+
         # (Perhaps) Initialize graphics.
         if not game.interface:
             renpy.display.core.Interface()
@@ -479,12 +618,19 @@ def main():
                 finally:
                     restart = (renpy.config.end_game_transition, "_invoke_main_menu", "_main_menu")
                     renpy.persistent.update(True)
+                    renpy.persistent.save_MP()
 
-            except game.FullRestartException, e:
+            except game.FullRestartException as e:
                 restart = e.reason
 
             finally:
 
+                # Reset live2d if it exists.
+                try:
+                    renpy.gl2.live2d.reset_states()
+                except:
+                    pass
+
                 # Flush any pending interface work.
                 renpy.display.interface.finish_pending()
 
@@ -493,6 +639,11 @@ def main():
 
     finally:
 
+        gc.set_debug(0)
+
+        for i in renpy.config.quit_callbacks:
+            i()
+
         renpy.loader.auto_quit()
         renpy.savelocation.quit()
         renpy.translation.write_updated_strings()
@@ -500,4 +651,3 @@ def main():
     # This is stuff we do on a normal, non-error return.
     if not renpy.display.error.error_handled:
         renpy.display.render.check_at_shutdown()
-
