@@ -1,4 +1,4 @@
-# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2020 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,7 +19,8 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
 
 import math
 import renpy.display
@@ -31,7 +32,9 @@ import renpy.text.texwrap as texwrap
 import renpy.text.font as font
 import renpy.text.extras as extras
 
-from _renpybidi import log2vis, WRTL, RTL, ON  # @UnresolvedImport
+from _renpybidi import log2vis, WRTL, RTL, ON # @UnresolvedImport
+
+BASELINE = -65536
 
 
 class Blit(object):
@@ -186,12 +189,14 @@ class TextSegment(object):
             self.ruby_bottom = source.ruby_bottom
             self.hinting = source.hinting
             self.outline_color = source.outline_color
+            self.ignore = source.ignore
 
         else:
             self.hyperlink = 0
             self.cps = 0
             self.ruby_top = False
             self.ruby_bottom = False
+            self.ignore = False
 
     def __repr__(self):
         return "<TextSegment font={font}, size={size}, bold={bold}, italic={italic}, underline={underline}, color={color}, black_color={black_color}, hyperlink={hyperlink}, vertical={vertical}>".format(**self.__dict__)
@@ -237,6 +242,9 @@ class TextSegment(object):
         Return the list of glyphs corresponding to unicode string s.
         """
 
+        if self.ignore:
+            return [ ]
+
         fo = font.get_font(self.font, self.size, self.bold, self.italic, 0, self.antialias, self.vertical, self.hinting, layout.oversample)
         rv = fo.glyphs(s)
 
@@ -291,13 +299,28 @@ class TextSegment(object):
         on the font group.
         """
 
-        if not isinstance(self.font, font.FontGroup):
-            yield (self, s)
+        tf = self.font
+
+        font_transform = renpy.game.preferences.font_transform
+        if font_transform is not None:
+            font_func = renpy.config.font_transforms.get(font_transform, None)
+            if font_func is not None:
+                tf = font_func(tf)
+
+        if not isinstance(tf, font.FontGroup):
+
+            if self.font is tf:
+                yield (self , s)
+            else:
+                seg = TextSegment(self)
+                seg.font = tf
+                yield (seg, s)
+
             return
 
         segs = { }
 
-        for f, ss in self.font.segment(s):
+        for f, ss in tf.segment(s):
 
             seg = segs.get(f, None)
 
@@ -337,7 +360,7 @@ class SpaceSegment(object):
         self.glyph = glyph = textsupport.Glyph()
 
         glyph.character = 0
-        glyph.ascent = 0
+        glyph.ascent = 1
         glyph.line_spacing = height
         glyph.advance = width
         glyph.width = width
@@ -355,6 +378,7 @@ class SpaceSegment(object):
 
     def draw(self, glyphs, di, xo, yo, layout):
         # Does nothing - since there's nothing to draw.
+
         return
 
     def assign_times(self, gt, glyphs):
@@ -383,6 +407,8 @@ class DisplayableSegment(object):
 
         self.hyperlink = ts.hyperlink
         self.cps = ts.cps
+        self.ruby_top = ts.ruby_top
+        self.ruby_bottom = ts.ruby_bottom
 
     def glyphs(self, s, layout):
 
@@ -400,7 +426,16 @@ class DisplayableSegment(object):
         if self.hyperlink:
             glyph.hyperlink = self.hyperlink
 
-        return [ glyph ]
+        rv = [ glyph ]
+
+        if self.ruby_bottom:
+            textsupport.mark_ruby_bottom(rv)
+        elif self.ruby_top == "alt":
+            textsupport.mark_altruby_top(rv)
+        elif self.ruby_top:
+            textsupport.mark_ruby_top(rv)
+
+        return rv
 
     def draw(self, glyphs, di, xo, yo, layout):
         glyph = glyphs[0]
@@ -465,6 +500,13 @@ class Layout(object):
             Layout (which must be another Layout of the same text).
         """
 
+        def find_baseline():
+            for g in all_glyphs:
+                if g.ascent:
+                    return g.y + self.yoffset
+
+            return 0
+
         width = min(32767, width)
         height = min(32767, height)
 
@@ -477,11 +519,18 @@ class Layout(object):
             self.reverse = renpy.display.draw.draw_to_virt
             self.forward = renpy.display.draw.virt_to_draw
 
+            self.outline_step = text.style.outline_scaling != "linear"
+
+            self.pixel_perfect = True
+
         else:
 
             self.oversample = 1.0
             self.reverse = renpy.display.render.IDENTITY
             self.forward = renpy.display.render.IDENTITY
+            self.outline_step = True
+
+            self.pixel_perfect = False
 
         style = text.style
 
@@ -602,7 +651,7 @@ class Layout(object):
             self.paragraph_glyphs.append(list(par_glyphs))
 
             if splits_from:
-                textsupport.copy_splits(splits_from.paragraph_glyphs[p_num], par_glyphs)  # @UndefinedVariable
+                textsupport.copy_splits(splits_from.paragraph_glyphs[p_num], par_glyphs) # @UndefinedVariable
 
             else:
 
@@ -697,6 +746,7 @@ class Layout(object):
         adjust_spacing = text.style.adjust_spacing
 
         if splits_from and adjust_spacing:
+
             target_x = self.scale_int(splits_from.size[0] - splits_from.xborder)
             target_y = self.scale_int(splits_from.size[1] - splits_from.yborder)
 
@@ -708,17 +758,21 @@ class Layout(object):
             elif adjust_spacing == "vertical":
                 target_x_delta = 0.0
 
-            textsupport.tweak_glyph_spacing(all_glyphs, lines, target_x_delta, target_y_delta, maxx, y)  # @UndefinedVariable
+            textsupport.tweak_glyph_spacing(all_glyphs, lines, target_x_delta, target_y_delta, maxx, y) # @UndefinedVariable
 
             maxx = target_x
             y = target_y
 
+            textsupport.offset_glyphs(all_glyphs, 0, int(round(splits_from.baseline * self.oversample)) - find_baseline())
+
         # Figure out the size of the texture. (This is a little over-sized,
         # but it simplifies the code to not have to care about borders on a
         # per-outline basis.)
         sw, sh = size = (maxx + self.xborder, y + self.yborder)
         self.size = size
 
+        self.baseline = find_baseline()
+
         # If we only care about the size, we're done.
         if size_only:
             return
@@ -773,6 +827,9 @@ class Layout(object):
 
                 ts.draw(glyphs, di, self.add_left, self.add_top, self)
 
+            if renpy.config.debug_text_alignment:
+                self.make_alignment_grid(surf)
+
             renpy.display.draw.mutated_surface(surf)
             tex = renpy.display.draw.load_texture(surf)
 
@@ -803,6 +860,22 @@ class Layout(object):
                 renpy.display.to_log.write("     Available: (%d, %d) Laid-out: (%d, %d)", width, height, sw, sh)
                 renpy.display.to_log.write("     Text: %r", text.text)
 
+    def make_alignment_grid(self, surf):
+        w, h = surf.get_size()
+
+        for x in range(w):
+            for y in range(h):
+
+                if surf.get_at((x, y))[3] > 0:
+                    continue
+
+                if x == 0 or y == 0 or x == w - 1 or y == h - 1:
+                    surf.set_at((x, y), (255, 0, 0, 255))
+                elif (x ^ y) & 1:
+                    surf.set_at((x, y), (255, 255, 255, 255))
+                else:
+                    surf.set_at((x, y), (0, 0, 0, 255))
+
     def scale(self, n):
         if n is None:
             return n
@@ -813,6 +886,9 @@ class Layout(object):
         if n is None:
             return n
 
+        if isinstance(n, renpy.display.core.absolute):
+            return int(n)
+
         return int(round(n * self.oversample))
 
     def scale_outline(self, n):
@@ -822,10 +898,25 @@ class Layout(object):
         if isinstance(n, renpy.display.core.absolute):
             return int(n)
 
-        if self.oversample < 1:
-            return n
+        if self.outline_step:
 
-        return n * int(self.oversample)
+            if self.oversample < 1:
+                return n
+
+            return n * int(self.oversample)
+
+        else:
+            if n == 0:
+                return 0
+
+            rv = round(n * self.oversample)
+
+            if n < 0 and rv > -1:
+                rv = -1
+            if n > 0 and rv < 1:
+                rv = 1
+
+            return rv
 
     def unscale_pair(self, x, y):
         return x / self.oversample, y / self.oversample
@@ -873,7 +964,7 @@ class Layout(object):
 
             line.extend(tss[-1].subsegment(u"\u200B"))
 
-        for type, text in tokens:  # @ReservedAssignment
+        for type, text in tokens: # @ReservedAssignment
 
             try:
 
@@ -954,6 +1045,9 @@ class Layout(object):
                 elif tag == "fast":
                     pass
 
+                elif tag == "done":
+                    pass
+
                 elif tag == "nw":
                     pass
 
@@ -1110,6 +1204,13 @@ class Layout(object):
                     ts = push()
                     ts.vertical = False
 
+                elif tag == "alt":
+                    ts = push()
+                    ts.ignore = True
+
+                elif tag == "noalt":
+                    ts = push()
+
                 elif tag[0] == "#":
                     pass
 
@@ -1140,7 +1241,7 @@ class Layout(object):
         l = [ ]
 
         for ts, s in p:
-            s, direction = log2vis(s, direction)
+            s, direction = log2vis(unicode(s), direction)
             l.append((ts, s))
 
         rtl = (direction == RTL or direction == WRTL)
@@ -1169,10 +1270,10 @@ class Layout(object):
                 dslist = [ dslist ]
 
             for dsx, dsy in dslist:
-                outlines.append((0, style.drop_shadow_color, self.scale_outline(dsx), self.scale_outline(dsy)))
+                outlines.append((0, style.drop_shadow_color, self.scale_int(dsx), self.scale_int(dsy)))
 
         for size, color, xo, yo in style_outlines:
-            outlines.append((self.scale_outline(size), color, self.scale_outline(xo), self.scale_outline(yo)))
+            outlines.append((self.scale_outline(size), color, self.scale_int(xo), self.scale_int(yo)))
 
         # The outline borders we reserve.
         left = 0
@@ -1264,13 +1365,13 @@ class Layout(object):
             if g.x + g.advance > max_x:
                 max_x = g.x + g.advance
 
-            if g.x  < min_x:
+            if g.x < min_x:
                 min_x = g.x
 
         ly = min(l.y + l.height + self.line_overlap_split, max_height)
 
         if min_x < max_x:
-            rv.append(Blit(min_x, max_y, max_x - min_x, ly - max_y, left=left, right=right, top=top, bottom=(l is self.lines[-1]) ))
+            rv.append(Blit(min_x, max_y, max_x - min_x, ly - max_y, left=left, right=right, top=top, bottom=(l is self.lines[-1])))
 
         return rv
 
@@ -1346,6 +1447,7 @@ VERT_FORWARD = renpy.display.render.Matrix2D(0, 1, -1, 0)
 class Text(renpy.display.core.Displayable):
 
     """
+    :name: Text
     :doc: text
     :args: (text, slow=None, scope=None, substitute=None, slow_done=None, **properties)
 
@@ -1542,10 +1644,8 @@ class Text(renpy.display.core.Displayable):
                     i, did_sub = renpy.substitutions.substitute(i, scope, substitute)
                     uses_scope = uses_scope or did_sub
 
-                if isinstance(i, str):
-                    i = unicode(i, "utf-8", "replace")
-                else:
-                    i = unicode(i)
+                if isinstance(i, bytes):
+                    i = str(i, "utf-8", "replace")
 
             new_text.append(i)
 
@@ -1630,7 +1730,7 @@ class Text(renpy.display.core.Displayable):
         # Tokenize the text.
         tokens = self.tokenize(text)
 
-        if renpy.config.custom_text_tags or (renpy.config.replace_text is not None):
+        if renpy.config.custom_text_tags or renpy.config.self_closing_custom_text_tags or (renpy.config.replace_text is not None):
             tokens = self.apply_custom_tags(tokens)
 
         # self.tokens is a list of pairs, where the first component of
@@ -1669,7 +1769,7 @@ class Text(renpy.display.core.Displayable):
         rv = "".join(rv)
         _, _, rv = rv.rpartition("{fast}")
 
-        rv = renpy.translation.dialogue.notags_filter(rv)
+        rv = renpy.text.extras.filter_alt_text(rv)
 
         alt = self.style.alt
 
@@ -1728,6 +1828,36 @@ class Text(renpy.display.core.Displayable):
 
         super(Text, self).set_style_prefix(prefix, root)
 
+    def get_placement(self):
+
+        rv = super(Text, self).get_placement()
+
+        if rv[3] != BASELINE:
+            return rv
+
+        layout = self.get_virtual_layout()
+
+        if layout is None:
+
+            width = 4096
+            height = 4096
+            st = 0
+            at = 0
+
+            if self.dirty or self.displayables is None:
+                self.update()
+
+            renders = { }
+
+            for i in self.displayables:
+                renders[i] = renpy.display.render.render(i, width, self.style.size, st, at)
+
+            layout = Layout(self, width, height, renders, size_only=True, drawable_res=True)
+
+        xpos, ypos, xanchor, yanchor, xoffset, yoffset, subpixel = rv
+        rv = (xpos, ypos, xanchor, layout.baseline, xoffset, yoffset, subpixel)
+        return rv
+
     def focus(self, default=False):
         """
         Called when a hyperlink gains focus.
@@ -1837,12 +1967,16 @@ class Text(renpy.display.core.Displayable):
 
     def size(self, width=4096, height=4096, st=0, at=0):
         """
+        :args: (width=4096, height=4096, st=0, at=0)
+
         Attempts to figure out the size of the text. The parameters are
         as for render.
 
         This does not rotate vertical text.
         """
 
+        # This is mostly duplicated in get_placement.
+
         if self.dirty or self.displayables is None:
             self.update()
 
@@ -1851,7 +1985,7 @@ class Text(renpy.display.core.Displayable):
         for i in self.displayables:
             renders[i] = renpy.display.render.render(i, width, self.style.size, st, at)
 
-        layout = Layout(self, width, height, renders, size_only=True)
+        layout = Layout(self, width, height, renders, size_only=True, drawable_res=True)
 
         return layout.unscale_pair(*layout.size)
 
@@ -2052,6 +2186,9 @@ class Text(renpy.display.core.Displayable):
             vrv.blit(rv, (rv.height, 0))
             rv = vrv
 
+        if layout.pixel_perfect:
+            rv.properties = { "pixel_perfect" : True }
+
         return rv
 
     def tokenize(self, text):
@@ -2063,11 +2200,11 @@ class Text(renpy.display.core.Displayable):
 
         for i in text:
 
-            if isinstance(i, unicode):
+            if isinstance(i, str):
                 tokens.extend(textsupport.tokenize(i))
 
-            elif isinstance(i, str):
-                tokens.extend(textsupport.tokenize(unicode(i)))
+            elif isinstance(i, basestring):
+                tokens.extend(textsupport.tokenize(str(i)))
 
             elif isinstance(i, renpy.display.core.Displayable):
                 tokens.append((DISPLAYABLE, i))
@@ -2090,7 +2227,7 @@ class Text(renpy.display.core.Displayable):
             kind, text = t
 
             if kind == TEXT and renpy.config.replace_text:
-                rv.append((TEXT, unicode(renpy.config.replace_text(text))))
+                rv.append((TEXT, str(renpy.config.replace_text(text))))
 
             elif kind != TAG:
                 rv.append(t)
@@ -2101,48 +2238,60 @@ class Text(renpy.display.core.Displayable):
 
                 func = renpy.config.custom_text_tags.get(tag, None)
 
+                if func is None:
+                    func = renpy.config.self_closing_custom_text_tags.get(tag, None)
+                    self_closing = True
+                else:
+                    self_closing = False
+
                 if func is None:
                     rv.append(t)
                     continue
 
-                # The contents of this tag.
-                contents = [ ]
+                if not self_closing:
 
-                # The close tag we're lookin for.
-                close = "/" + tag
+                    # The contents of this tag.
+                    contents = [ ]
 
-                # The number of open tags.
-                count = 1
+                    # The close tag we're lookin for.
+                    close = "/" + tag
 
-                while tokens:
+                    # The number of open tags.
+                    count = 1
 
-                    # Count the number of `tag` tags that are still open.
-                    t2 = tokens.pop(0)
+                    while tokens:
 
-                    kind2, text2 = t2
+                        # Count the number of `tag` tags that are still open.
+                        t2 = tokens.pop(0)
 
-                    if kind2 == TAG:
-                        tag2, _, _ = text2.partition("=")
+                        kind2, text2 = t2
 
-                        if tag2 == tag:
-                            count += 1
-                        elif tag2 == close:
-                            count -= 1
-                            if not count:
-                                break
+                        if kind2 == TAG:
+                            tag2, _, _ = text2.partition("=")
 
-                    contents.append(t2)
+                            if tag2 == tag:
+                                count += 1
+                            elif tag2 == close:
+                                count -= 1
+                                if not count:
+                                    break
 
-                if count:
-                    raise Exception("Text ended while the '{}' text tag was still open.".format(tag))
+                        contents.append(t2)
 
-                new_contents = func(tag, value, contents)
+                    if count:
+                        raise Exception("Text ended while the '{}' text tag was still open.".format(tag))
+
+                    new_contents = func(tag, value, contents)
+
+                else:
+
+                    new_contents = func(tag, value)
 
                 new_tokens = [ ]
 
                 for kind2, text2 in new_contents:
-                    if isinstance(text2, str):
-                        text2 = unicode(text2)
+                    if isinstance(text2, bytes):
+                        text2 = str(text2)
 
                     new_tokens.append((kind2, text2))
 
